/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef __PIKA_PALETTE_H__
#define __PIKA_PALETTE_H__


#include "pikadata.h"


#define PIKA_TYPE_PALETTE            (pika_palette_get_type ())
#define PIKA_PALETTE(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), PIKA_TYPE_PALETTE, PikaPalette))
#define PIKA_PALETTE_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), PIKA_TYPE_PALETTE, PikaPaletteClass))
#define PIKA_IS_PALETTE(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), PIKA_TYPE_PALETTE))
#define PIKA_IS_PALETTE_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), PIKA_TYPE_PALETTE))
#define PIKA_PALETTE_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), PIKA_TYPE_PALETTE, PikaPaletteClass))


struct _PikaPaletteEntry
{
  PikaRGB  color;
  gchar   *name;
};


typedef struct _PikaPaletteClass PikaPaletteClass;

struct _PikaPalette
{
  PikaData  parent_instance;

  GList    *colors;
  gint      n_colors;

  gint      n_columns;
};

struct _PikaPaletteClass
{
  PikaDataClass  parent_class;

  void (* entry_changed) (PikaPalette *palette,
                          gint         index);
};


GType              pika_palette_get_type        (void) G_GNUC_CONST;

PikaData         * pika_palette_new             (PikaContext      *context,
                                                 const gchar      *name);
PikaData         * pika_palette_get_standard    (PikaContext      *context);

GList            * pika_palette_get_colors      (PikaPalette      *palette);
gint               pika_palette_get_n_colors    (PikaPalette      *palette);

void               pika_palette_move_entry      (PikaPalette      *palette,
                                                 PikaPaletteEntry *entry,
                                                 gint              position);

PikaPaletteEntry * pika_palette_add_entry       (PikaPalette      *palette,
                                                 gint              position,
                                                 const gchar      *name,
                                                 const PikaRGB    *color);
void               pika_palette_delete_entry    (PikaPalette      *palette,
                                                 PikaPaletteEntry *entry);

gboolean           pika_palette_set_entry       (PikaPalette      *palette,
                                                 gint              position,
                                                 const gchar      *name,
                                                 const PikaRGB    *color);
gboolean           pika_palette_set_entry_color (PikaPalette      *palette,
                                                 gint              position,
                                                 const PikaRGB    *color,
                                                 gboolean          push_undo_if_image);
gboolean           pika_palette_set_entry_name  (PikaPalette      *palette,
                                                 gint              position,
                                                 const gchar      *name);
PikaPaletteEntry * pika_palette_get_entry       (PikaPalette      *palette,
                                                 gint              position);
gint               pika_palette_get_entry_position (PikaPalette   *palette,
                                                 PikaPaletteEntry *entry);

void               pika_palette_set_columns     (PikaPalette      *palette,
                                                 gint              columns);
gint               pika_palette_get_columns     (PikaPalette      *palette);

PikaPaletteEntry * pika_palette_find_entry      (PikaPalette      *palette,
                                                 const PikaRGB    *color,
                                                 PikaPaletteEntry *start_from);


#endif /* __PIKA_PALETTE_H__ */
